#include <assert.h>
#include "libcomp.h"



// flop = multiply-add

inline void DoFlops(const unsigned flopsperelement)
{
  static double sum;
  static volatile double x;
  
  double val;
  double mpy;

  sum=x;
  for (unsigned i=0;i<flopsperelement;i++) {
    val=x;
    mpy=x;
    sum = sum + mpy*val;
  }
}


static volatile double mem[MAXMEMPERELEMENT];

inline void DoReads(const unsigned readsperelement) 
{
  static double sum;

  for (unsigned i=0;i<readsperelement;i++) {
    sum=mem[i];
  }
}


inline void DoWrites(const unsigned writesperelement) 
{
  static double sum;
  static volatile double x;

  sum=x;

  for (unsigned i=0;i<writesperelement; i++) {
    mem[i]=sum;
  }
}
  



void RunIter(unsigned          numelements,
	     const unsigned    flopsperelement,
	     const unsigned    readsperelement,
	     const unsigned    writesperelement)
{
  assert(numelements>0);
  assert(flopsperelement<=MAXFLOPSPERELEMENT);
  assert(readsperelement<=MAXMEMPERELEMENT);
  assert(writesperelement<=MAXMEMPERELEMENT);

  for (unsigned i=0;i<numelements;i++) { 
    if (flopsperelement>0) {DoFlops(flopsperelement);}
    if (readsperelement>0) {DoReads(readsperelement);}
    if (flopsperelement>0) {DoWrites(writesperelement);}
  }
}



Time::Time(const Time &rhs) : timeval((timeval&)rhs)
{}

Time::Time(const timeval &rhs) : timeval(rhs)
{}

Time::Time(const double time) 
{
  tv_sec = (int)time;
  tv_usec= (int)((time-((double)((int)time)))*1e6);
}

Time::Time(const unsigned sec, const unsigned usec) 
{
  tv_sec=sec;
  tv_usec=usec;
}
    
Time::Time()
{
  SetToCurrentTime();
}

void Time::SetToCurrentTime()
{
  gettimeofday((struct timeval *)this,0);
}

Time & Time::operator=(const Time &rhs)
{
  tv_sec=rhs.tv_sec;
  tv_usec=rhs.tv_usec;
  return *this;
}

Time & Time::operator=(const double &rhs)
{
  return operator=(Time(rhs));
}

Time::operator double() const
{
  return ((double)(tv_sec)) + ((double)(tv_usec)/1e6);
}

bool Time::operator<(const Time &rhs) const
{
  if (tv_sec<rhs.tv_sec) { 
    return true;
  } else {
    if (tv_sec>rhs.tv_sec) { 
      return false;
    } else {
      return (tv_usec<rhs.tv_usec);
    }
  }
}

bool Time::operator>(const Time &rhs) const
{
  return !(operator <(rhs));
}

bool Time::operator==(const Time &rhs) const
{
  return tv_sec==rhs.tv_sec && tv_usec==rhs.tv_usec;
}

ostream & Time::Print(ostream &os) const
{
  os << "Time(sec="<<tv_sec<<", usec="<<tv_usec<<", double="<<((double)(*this))<<")";
  return os;
}

  
  
