#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <math.h>
#include "libcomp.h"


#include "pvm3.h"
#include "pvmspawn.h"


#include <vector>
#include <algorithm>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>


#define INIT    1
#define INITOK  2

#define TIMINGS    16
#define RESULTS    32

#define ITERBASE 100

int      *tids;
int       numprocs;



enum Type {SEND,RECEIVE};

struct Action
{
  Type type;
  int node;
  Action(const Type &t, const int n) : type(t), node(n)
  {}
}
;

struct Schedule
{
  int thisnode;
  vector<Action> actions;
};


void PrintSchedule(const Schedule *sched)
{
  cout << "Schedule for node "<<sched->thisnode<<endl;
  for (vector<Action>::const_iterator step=sched->
	 actions.begin();
       step!=sched->actions.end();
       step++)
    {
      cout << ((*step).type==SEND ? "SEND" : "RECEIVE") <<" "<<(*step).node<<endl;
    }
}

bool IsPow2(unsigned n)
{
  int numbits=0;
  while (n)
    {
      numbits+=n&0x1;
      n>>=1;
    }
  return numbits==1;
}



unsigned Log2(unsigned n)
{
  int pow=1;
  while (n)
    {
      pow++;
      n>>=1;
    }
  return pow;
}

unsigned Pow2(unsigned n)
{
  return 0x1<<n;
}

int min(int a,int b)
{
  return (a<b) ? a : b;
}

Schedule *compute_schedule_mesh_toroid_neighbor1d(int numnodes, int thisnode,bool toroid)
{
  bool even = !(thisnode % 2);
  int nextnode=thisnode+1;
  if (toroid)
    {
      nextnode%=numnodes;
    }
  int prevnode=thisnode-1;
  if (toroid && prevnode<0)
    {
      prevnode=numnodes-1;
    }

  Schedule *s = new Schedule;
  s->thisnode=thisnode;
  if (even)
    {
      if (nextnode<numnodes)
        {
	  s->actions.push_back(Action(SEND,nextnode));
	  s->actions.push_back(Action(RECEIVE,nextnode));
        }
      if (prevnode>=0)
        {
	  s->actions.push_back(Action(RECEIVE,prevnode));
	  s->actions.push_back(Action(SEND,prevnode));
        }
    }
  if (!even)
    {
      if (prevnode>=0)
        {
	  s->actions.push_back(Action(RECEIVE,prevnode));
	  s->actions.push_back(Action(SEND,prevnode));
        }
      if (nextnode<numnodes)
        {
	  s->actions.push_back(Action(SEND,nextnode));
	  s->actions.push_back(Action(RECEIVE,nextnode));
        }
    }
  return s;
}


Schedule *compute_schedule_mesh_neighbor1d(int numnodes, int thisnode)
{
  return compute_schedule_mesh_toroid_neighbor1d(numnodes,thisnode,false);
}

Schedule *compute_schedule_mesh_toroid_neighbor2d(int numnodes, int thisnode,bool toroid)
{
  int i;
  bool even = !(thisnode % 2);

  int w=ceil(sqrt(numnodes));

  for(i=w;i<=numnodes;i++)
    {
      if(numnodes % i==0)
        {
	  w=i;
	  break;
        }
    }
  int h=numnodes/w;

  bool roweven= !((thisnode/w)%2);

  int left=(thisnode/w)*w;
  int right=left+w-1;


  int nextnode=thisnode+1;
  if(nextnode>right)
    {
      if (toroid)
        {
	  nextnode-=w;
        }
      else
        {
	  nextnode=numnodes; // no communication
        }
    }

  int prevnode=thisnode-1;
  if(prevnode<left)
    {
      if (toroid)
        {
	  prevnode+=w;
        }
      else
        {
	  prevnode=-1;    // no communication
        }

    }


  int topnode=thisnode-w;
  if(topnode<0)
    {
      if (toroid)
        {
	  topnode+=numnodes;
        }
      else
        {
	  topnode=-1;
        }
    }

  int bottomnode=thisnode+w;
  if(bottomnode>=numnodes)
    {
      if (toroid)
        {
	  bottomnode-=numnodes;
        }
      else
        {
	  bottomnode=numnodes;
        }
    }


  Schedule *s = new Schedule;
  s->thisnode=thisnode;
  if (even)
    {
      if (nextnode<numnodes)
        {
	  s->actions.push_back(Action(SEND,nextnode));
	  s->actions.push_back(Action(RECEIVE,nextnode));
        }
      if (prevnode>=0)
        {
	  s->actions.push_back(Action(RECEIVE,prevnode));
	  s->actions.push_back(Action(SEND,prevnode));
        }
    }

  if (!even)
    {
      if (prevnode>=0)
        {
	  s->actions.push_back(Action(RECEIVE,prevnode));
	  s->actions.push_back(Action(SEND,prevnode));
        }
      if (nextnode<numnodes)
        {
	  s->actions.push_back(Action(SEND,nextnode));
	  s->actions.push_back(Action(RECEIVE,nextnode));
        }
    }


  if(roweven)
    {
      if (bottomnode<numnodes)
        {
	  s->actions.push_back(Action(SEND,bottomnode));
	  s->actions.push_back(Action(RECEIVE,bottomnode));
        }
      if (topnode>=0 && topnode!=bottomnode)
        {
	  s->actions.push_back(Action(RECEIVE,topnode));
	  s->actions.push_back(Action(SEND,topnode));
        }


    }
  else
    {
      if (topnode>=0)
        {
	  s->actions.push_back(Action(RECEIVE,topnode));
	  s->actions.push_back(Action(SEND,topnode));
        }

      if (bottomnode<numnodes && bottomnode!=topnode)
        {
	  s->actions.push_back(Action(SEND,bottomnode));
	  s->actions.push_back(Action(RECEIVE,bottomnode));
        }
    }
  return s;
}

bool sortaction(Action a1,Action a2)
{
  if(a1.node<a2.node)
    return true;
  return false;

}

Schedule *compute_schedule_mesh_toroid_neighbor(int config, int thisnode,bool toroid)
{
  int dim[5],dimm[6];
  int nd;
  int numnodes;

  // find the config from the config variable
  int i,j;
  dim[0]=1;
  for(i=1;i<5;i++)
    {
      dim[i]=config%100;
      config/=100;
      if(config==0)
	break;
    }
  nd=i; // now we know the number of dimensions

  // set up the dimm arrays
  dimm[0]=1;
  for(i=1;i<=nd;i++)
    {
      dimm[i]=dimm[i-1]*dim[i];
    }
  numnodes=dimm[nd];

  // print the dim and dimm arrays
  cout << "\n\ndim\tdimm" << endl;
  for(i=0;i<=nd;i++)
    {
      cout << dim[i] << "\t" << dimm[i] << endl;

    }

  Schedule *s = new Schedule;
  s->thisnode=thisnode;

  // run the for loop , including more first and last nodes to have the edge bindings ( needed for the toroid )
  for(i=0-dimm[nd-1];i<=numnodes-1+dimm[nd-1];i++)
    {

      int nb=i;
      int diff=nb-thisnode;
      // run a loop for each of the possible neighbor differences

      for(j=0;j<nd;j++)
        {
	  if(abs(diff)==dimm[j]) // there is an edge between i and thisnode
            {
	      // compute the bounds
	      int b1=(thisnode/dimm[j+1])*dimm[j+1];
	      int b2=b1+dimm[j+1]-1;

	      if(nb<b1) // out of b1
                {
		  if(toroid)
                    {
		      nb+=dimm[j+1];
		      if(dimm[j+1]==2*dimm[j])
			nb=-1; // collision !
                    }
		  else     //discard
                    {
		      nb=-1;
                    }
                }

	      if(nb>b2) // out of b2
                {
		  if(toroid)
                    {
		      nb-=dimm[j+1];
		      if(dimm[j+1]==2*dimm[j])
			nb=numnodes; // collision !
                    }
		  else     //discard
                    {
		      nb=numnodes;
                    }
                }

	      // there is a definite edge between nb and thisnode
	      bool positive=(nb>thisnode) ? true : false;
	      if(positive && nb>=0 && nb<numnodes)
                {
		  s->actions.push_back(Action(SEND,nb));
		  s->actions.push_back(Action(RECEIVE,nb));

                }
	      if(!positive && nb>=0 && nb<numnodes)
                {
		  s->actions.push_back(Action(RECEIVE,nb));
		  s->actions.push_back(Action(SEND,nb));
                }

            }     // there is an edge between i and thisnode
        }
      // is there an edge

    }
  // include filters in it, to have the toroid special edges
  // to exclude collisions, make sure the filter is not twice the dimension value
  sort(s->actions.begin(),s->actions.end(),sortaction);

  return s;


}

Schedule *compute_schedule_mesh_toroid_neighbor3d(int numnodes, int thisnode,bool toroid)
{
  return 0;
}


Schedule *compute_schedule_mesh_neighbor2d(int numnodes, int thisnode)
{
  return compute_schedule_mesh_toroid_neighbor2d(numnodes,thisnode,false);
}

Schedule *compute_schedule_mesh_neighbor3d(int numnodes, int thisnode)
{
  return 0;
}
Schedule *compute_schedule_toroid_neighbor1d(int numnodes, int thisnode)
{
  return compute_schedule_mesh_toroid_neighbor1d(numnodes,thisnode,true);
}
Schedule *compute_schedule_toroid_neighbor2d(int numnodes, int thisnode)
{
  return compute_schedule_mesh_toroid_neighbor2d(numnodes,thisnode,true);
}
Schedule *compute_schedule_toroid_neighbor3d(int numnodes, int thisnode)
{
  return 0;
}
Schedule *compute_schedule_reduction_tree(int numnodes, int thisnode)
{


  Schedule *s = new Schedule;
  s->thisnode=thisnode;

  // binary reduction tree followed by broadcast
  // recieve from children;
  int c1=2*thisnode+1;
  int c2=2*thisnode+2;
  if(c1<numnodes)
    s->actions.push_back(Action(RECEIVE,c1));
  if(c2<numnodes)
    s->actions.push_back(Action(RECEIVE,c2));

  // send to parent
  int p=(thisnode-1)/2;
  if(p!=thisnode)
    s->actions.push_back(Action(SEND,p));

  // Should just borrow this from the SCANMACS code
  //

  /*
    if (!IsPow2(numnodes)) {
    cout << "Can't schedule reduction tree.  Not a power of two nodes.\n";
    return 0;
    }

    unsigned numlevels=Log2(numnodes);

    // upward reduction
    unsigned i;
    for (i=1;i<=numlevels;i++) {
    unsigned p2=Pow2(i);
    if (!(thisnode % p2)) {
    if (

  */


  return s;
}

bool is_pow2(int x)
{
  int i;
  for(i=0;i<=(log(x)/log(2)+1);i++)
    {
      if(x==pow(2,i))
	return true;
    }

  return false;
}

Schedule *compute_schedule_hypercube_neighbor(int numnodes, int thisnode)
{
  int i;

  Schedule *s = new Schedule;
  s->thisnode=thisnode;

  for(i=0;i<numnodes;i++)
    {
      int diff=i-thisnode;
      int xr=i ^ thisnode;


      if(!is_pow2(xr))
	continue;  // not hypercube neigbors
      //        if(abs(diff)==min(i,thisnode))
      //            continue; // reject this one
      //        if(abs(diff)==1 && min(i,thisnode)!=1)
      //            continue;

      if(diff==0)
	continue;

      bool positive= (diff>0);

      if(positive)
        {
	  s->actions.push_back(Action(SEND,i));
	  s->actions.push_back(Action(RECEIVE,i));
        }
      if(!positive)
        {
	  s->actions.push_back(Action(RECEIVE,i));
	  s->actions.push_back(Action(SEND,i));
        }

    }
  return s;
}
Schedule *compute_schedule_all_to_all(int numnodes, int thisnode)
{
  int i;

  Schedule *s = new Schedule;
  s->thisnode=thisnode;

  for(i=0;i<numnodes;i++)
    {
      int diff=i-thisnode;
      if(diff==0)
	continue;
      bool even=!(diff%2);
      bool positive= (diff>0);

      if(!even and positive)
        {
	  s->actions.push_back(Action(SEND,i));
	  s->actions.push_back(Action(RECEIVE,i));
        }
      if(!even and !positive)
        {
	  s->actions.push_back(Action(RECEIVE,i));
	  s->actions.push_back(Action(SEND,i));
        }

      if(even and positive)
        {
	  s->actions.push_back(Action(SEND,i));
	  s->actions.push_back(Action(RECEIVE,i));
        }
      if(even and !positive)
        {
	  s->actions.push_back(Action(RECEIVE,i));
	  s->actions.push_back(Action(SEND,i));
        }
    }
  return s;
}

Schedule *compute_schedule(char *pattern, int numnodes, int thisnode)
{
  if (!strcasecmp(pattern,"mesh_neighbor-1d"))            // DONE
    {
      return compute_schedule_mesh_toroid_neighbor1d(numnodes,thisnode,false);
    }
  else if (!strcasecmp(pattern,"mesh_neighbor-2d"))      // DONE
    {
      return compute_schedule_mesh_toroid_neighbor2d(numnodes,thisnode,false);
    }
  else if (!strcasecmp(pattern,"mesh_neighbor-3d"))
    {
      return compute_schedule_mesh_neighbor3d(numnodes,thisnode);
    }
  else if (!strcasecmp(pattern,"toroid_neighbor-1d"))        // DONE
    {
      return compute_schedule_mesh_toroid_neighbor1d(numnodes,thisnode,true);
    }
  else if (!strcasecmp(pattern,"toroid_neighbor-2d"))      // DONE
    {
      return compute_schedule_mesh_toroid_neighbor2d(numnodes,thisnode,true);
    }
  else if (!strcasecmp(pattern,"toroid_neighbor-3d"))
    {
      return compute_schedule_mesh_toroid_neighbor3d(numnodes,thisnode,true);
    }
  else if (!strcasecmp(pattern,"reduction_tree"))      //done
    {
      return compute_schedule_reduction_tree(numnodes,thisnode);
    }
  else if (!strcasecmp(pattern,"hypercube_neighbor"))    //done
    {
      return compute_schedule_hypercube_neighbor(numnodes,thisnode);
    }
  else if (!strcasecmp(pattern,"all-to-all"))           // done
    {
      return compute_schedule_all_to_all(numnodes,thisnode);
    }
  else if (!strncasecmp(pattern,"mesh",4))           // done
    {
      int config=atoi(&pattern[4]);
      return  compute_schedule_mesh_toroid_neighbor(config, thisnode,false);

    }
  else if (!strncasecmp(pattern,"toroid",6))           // done
    {
      int config=atoi(&pattern[6]);
      return  compute_schedule_mesh_toroid_neighbor(config, thisnode,true);

    }
  else
    {
      return 0;
    }
}



void usage()
{
  cerr << "usage: pattern pat numprocs messagesize numiters numelements flopsperelement readsperelement writesperelement\n";
}



int tid2pid(int tid)
{
  for (int i=0;i<numprocs;i++)
    {
      if (tids[i]==tid)
        {
	  return i;
        }
    }
  return -1;
}

int pid2tid(int pid)
{
  if (pid>=0 && pid<numprocs)
    {
      return tids[pid];
    }
  else
    {
      return -1;
    }
}


unsigned ceildiv(unsigned top, unsigned bot)
{
  return top/bot + (top%bot != 0);
}

unsigned floordiv(unsigned top, unsigned bot)
{
  return top/bot;
}


struct DANCE_STEPS
{
  char pattern[255]; // the name of the pattern
  int numprocs;
  unsigned messagesize;
  unsigned numiters;
  unsigned numelements;
  unsigned flopsperelement;
  unsigned readsperelement;
  unsigned writesperelement;
} dance_steps[128];

int pattern_count=1;
int main(int argc, char *argv[])
{
  unsigned numelements;
  unsigned messagesize;
  unsigned numiters;
  unsigned flopsperelement;
  unsigned readsperelement;
  unsigned writesperelement;
  char pattern[1024];
  int rc;

  int mytid = pvm_mytid();

  if (mytid<0)
    {
      cerr << "PVM has not been started.  pattern cannot run.\n";
      exit(-1);
    }

  Time start1;
  pvm_setopt(PvmRoute,PvmRouteDirect); // direct TCP Communication

  int myparent = pvm_parent();
  int myprocid;

  if (myparent==PvmNoParent)
    {
      if (argc!=9 && argc!=2)
        {
	  usage();
	  return -1;
        }

      if(argc==9)
	{
	  strcpy(pattern,argv[1]);
	  numprocs=atoi(argv[2]);
	  messagesize=atoi(argv[3]);
	  numiters=atoi(argv[4]);
	  numelements=atoi(argv[5]);
	  flopsperelement=atoi(argv[6]);
	  readsperelement=atoi(argv[7]);
	  writesperelement=atoi(argv[8]);

	  cout << "I am node zero and I think the world looks like this:"<<endl;
	  cout << " pattern      = "<<pattern<<endl;
	  cout << " numprocs     = "<<numprocs<<endl;
	  cout << " messageize   = "<<messagesize<<endl;
	  cout << " numiters     = "<<numiters<<endl;
	  cout << " numelements  = "<<numelements<<endl;
	  cout << " flopsperelement  = "<<flopsperelement<<endl;
	  cout << " readsperelement  = "<<readsperelement<<endl;
	  cout << " writesperelement = "<<writesperelement<<endl;

	  // See if we can compute a pattern

	  Schedule *sched = compute_schedule(pattern,numprocs,0);

	  if (!sched)
	    {
	      cout << "Can't compute schedule for pattern.\n";
	      exit(-1);
	    }
	  PrintSchedule(sched);
	  delete sched;
	
	}
      else // argc=2 , i.e. read from a file
	{
	  // now we have a bunch of patterns... need to parse this file and get them
	    
	  fstream f;
	  f.open(argv[1],ios::in);
	  if(!f)
	    {
	      cerr << "Error opening file " << argv[1] << " for input." << endl;
	      exit(-1);
	    }

	  pattern_count=0;
	  numprocs=0;
	  while(!f.eof())
	    {
	      f >> dance_steps[pattern_count].pattern >> dance_steps[pattern_count].numprocs >> dance_steps[pattern_count].messagesize >> dance_steps[pattern_count].numiters >> dance_steps[pattern_count].numelements >> dance_steps[pattern_count].flopsperelement >> dance_steps[pattern_count].readsperelement >> dance_steps[pattern_count].writesperelement;
		
	      if(dance_steps[pattern_count].numprocs==0) break;

	      if(dance_steps[pattern_count].numprocs>numprocs)
		numprocs=dance_steps[pattern_count].numprocs;
	      
	      cout << "Reading in a pattern : " << dance_steps[pattern_count].pattern << " , message size: " << dance_steps[pattern_count].messagesize << " , numiters : " << dance_steps[pattern_count].numiters << endl;

	      // test this pattern if it makes sense
	      Schedule *sched = compute_schedule(dance_steps[pattern_count].pattern,dance_steps[pattern_count].numprocs,0);
	    
	      if (!sched)
		{
		  cout << "Can't compute schedule for pattern.\n";
		  exit(-1);
		}
	      PrintSchedule(sched);
	      delete sched;
	    
	      pattern_count++;
	    }; // while loop
	  f.close();
	} // if argc==2

      // now we have an array of schedules !!!
	
      myprocid=0;
	
      tids = new int[numprocs];
      tids[myprocid]=mytid;
	
      if (pvm_joingroup("barrier")!=0)
	{
	  cout << "Can't join group or group already created."<<endl;
	  pvm_exit();
	  return -1;
	}
	
      cout << "Spawning processes now"<<endl;
	
      cout << "parameters to spawnproc : " << argv[0] << "," << numprocs-1 << endl;

      rc=spawnprocs(argv[0],0,numprocs-1,tids);

      if (rc!=numprocs-1)
        {
	  cout << "Failed to spawn processes correctly. rc="<<rc<<",  killing and quiting"<<endl;
	  if (rc>0)
            {
	      for (int i=0;i<rc;i++)
                {
		  pvm_kill(tids[i+1]);
                }
            }
	  pvm_lvgroup("barrier");
	  pvm_exit();
	  return -1;
        }

      Time end1;
	
      cout <<"Time taken for adding processes is : " << endl;
      cout << double(end1)-double(start1) << endl;

      cout << "Processes spawned, doing intial data exchange"<<endl;
    }
  else
    {
      cout << "Hello, my TID is "<<mytid<<" and I am starting to initialize."<<endl;
      pvm_joingroup("barrier");
      pattern_count=1;
    }
  
  Time starttime;
  
  for(int k=0;k<pattern_count;k++)
    {
      if (myparent==PvmNoParent)
	{

	  // the initialization phase for both the parent and the other guys
 
	  cout << "\nThis is node 0 : Preparing for dancestep : " << k << endl;
	  cout << "Description : Pattern : " << dance_steps[k].pattern << endl;
	  pvm_initsend(PvmDataInPlace);  // no copy
	  pvm_pkint(&pattern_count,1,1);
	  pvm_pkbyte(dance_steps[k].pattern,1024,1);
	  pvm_pkint(&numprocs,1,1); // the max number of procs
	  pvm_pkint(&dance_steps[k].numprocs,1,1); // number of proces for this dancestep
	  pvm_pkuint(&dance_steps[k].messagesize,1,1);
	  pvm_pkuint(&dance_steps[k].numiters,1,1);
	  pvm_pkuint(&dance_steps[k].numelements,1,1);
	  pvm_pkuint(&dance_steps[k].flopsperelement,1,1);
	  pvm_pkuint(&dance_steps[k].readsperelement,1,1);
	  pvm_pkuint(&dance_steps[k].writesperelement,1,1);
	  pvm_pkint(tids,numprocs,1);
	  for (int i=1;i<numprocs; i++ )
	    {
	      pvm_send(pid2tid(i),INIT);
	    }

	  int numok=1;

	  cout << tid2pid(mytid) << " (tid="<<mytid<<") is OK" << endl;
	  while (numok<numprocs)
	    {
	      int bufid=pvm_recv(-1,INITOK);
	      int bytes,tag,tid;
	      pvm_bufinfo(bufid,&bytes,&tag,&tid);
	      cout << tid2pid(tid) << " (tid="<<tid<<") is OK" << endl;
	      numok++;
	    }
 
	  cout << "Initialization for dance step : " << k << " is finished.  Beginning to compute."<<endl;	 
	}
      else
	{
      
	  pvm_recv(myparent,INIT);
	  pvm_upkint(&pattern_count,1,1);
	  pvm_upkbyte(dance_steps[k].pattern,1024,1);
	  pvm_upkint(&numprocs,1,1);
	  pvm_upkint(&dance_steps[k].numprocs,1,1);
	  pvm_upkuint(&dance_steps[k].messagesize,1,1);
	  pvm_upkuint(&dance_steps[k].numiters,1,1);
	  pvm_upkuint(&dance_steps[k].numelements,1,1);
	  pvm_upkuint(&dance_steps[k].flopsperelement,1,1);
	  pvm_upkuint(&dance_steps[k].readsperelement,1,1);
	  pvm_upkuint(&dance_steps[k].writesperelement,1,1);
	  
	  tids = new int [numprocs];

	  pvm_upkint(tids,numprocs,1);

	  myprocid=tid2pid(mytid);

	  cout << "I am node "<<myprocid<<" and I think the world looks like this:"<<endl;
	  cout << " pattern      = "<<dance_steps[k].pattern<<endl;
	  cout << " numprocs(local)     = "<<dance_steps[k].numprocs<<endl;
	  cout << " numprocs(global)     = "<< numprocs<<endl;
	  cout << " messageize   = "<<dance_steps[k].messagesize<<endl;
	  cout << " numiters     = "<<dance_steps[k].numiters<<endl;
	  cout << " numelements  = "<<dance_steps[k].numelements<<endl;
	  cout << " flopsperelement  = "<<dance_steps[k].flopsperelement<<endl;
	  cout << " readsperelement  = "<<dance_steps[k].readsperelement<<endl;
	  cout << " writesperelement = "<<dance_steps[k].writesperelement<<endl;
	  pvm_initsend(PvmDataInPlace);
	  pvm_send(myparent,INITOK);

	  if(myprocid>=dance_steps[k].numprocs) continue; // other nodes need not do anything !!

	}


      
      Schedule *sched=compute_schedule(dance_steps[k].pattern,dance_steps[k].numprocs,myprocid);

      PrintSchedule(sched);

      char *receive_mem=new char[dance_steps[k].messagesize];
      char *send_mem=new char[dance_steps[k].messagesize];


      pvm_barrier("barrier",dance_steps[k].numprocs);

      struct rusage rusagein, rusageout;

      getrusage(RUSAGE_SELF,&rusagein);

      Time start;


      double totaltime=0;
      double computetime=0;
      double* timeperiter=new double[dance_steps[k].numiters];


      unsigned i;
      for ( i=0;i<dance_steps[k].numiters;i++)
	{
	  Time time1;
	  for (vector<Action>::const_iterator step=sched->
		 actions.begin();
	       step!=sched->actions.end();
	       step++)
	    {
	      if ((*step).type==SEND)
		{
		  //cout << "I am " << myprocid << " : sending to " << (*step).node << endl;

		  pvm_psend(pid2tid((*step).node),ITERBASE+i,send_mem,dance_steps[k].messagesize,PVM_BYTE);
		}
	      else if ((*step).type==RECEIVE)
		{
		  //cout << "I am " << myprocid << " : receiving from " << (*step).node << endl;
		  pvm_precv(pid2tid((*step).node),ITERBASE+i,receive_mem,dance_steps[k].messagesize,PVM_BYTE,0,0,0);
		}
	    }
	  Time time3;

	  RunIter(dance_steps[k].numelements,dance_steps[k].flopsperelement,dance_steps[k].readsperelement,dance_steps[k].writesperelement);

	  // one iteration finsished !
	  Time time2;

	  // capture the average here
	  timeperiter[i]=double(time2)-double(time1);
	  computetime+=(double(time2)-double(time3));


	  totaltime+=timeperiter[i];
	}

      double iter_avg=totaltime/dance_steps[k].numiters;
    
      // calc stddev
      double stdev=0;
      for(i=0;i<dance_steps[k].numiters;i++)
	{
	  stdev+=(timeperiter[i]-iter_avg)*(timeperiter[i]-iter_avg);
	}
    
      stdev=sqrt(stdev/(dance_steps[k].numiters-1));

      pvm_barrier("barrier",dance_steps[k].numprocs);

      Time end;
      getrusage(RUSAGE_SELF,&rusageout);

      double wallclock=(double)end - (double)start;

      double usrtime = (double)Time(rusageout.ru_utime) - (double)Time(rusagein.ru_utime);
      double systime = (double)Time(rusageout.ru_stime) - (double)Time(rusagein.ru_stime);

      if (myprocid > 0)
	{
	  // return the timing data
	  pvm_initsend(PvmDataInPlace);
	  pvm_pkdouble(&wallclock,1,1);
	  pvm_pkdouble(&usrtime,1,1);
	  pvm_pkdouble(&systime,1,1);
	  pvm_send(myparent,TIMINGS);

	  // quit

	}
      else
	{

	  cout << "Communication:"<<endl;
	  for (int i=0;i<dance_steps[k].numprocs;i++)
	    {
	      if (i==0)
		{
		  // wallclock, usrtime, systime as above.
		}
	      else
		{
		  pvm_recv(pid2tid(i),TIMINGS);
		  pvm_upkdouble(&wallclock,1,1);
		  pvm_upkdouble(&usrtime,1,1);
		  pvm_upkdouble(&systime,1,1);
		}
	      cout << "Process "<<i<<":"<<endl;
	      cout << " Wallclock          = " << wallclock << " seconds"<<endl;
	      cout << " User               = " << usrtime << " seconds"<<endl;
	      cout << " System             = " << systime << " seconds"<<endl;
	      cout << " Average_per_iter   = " << iter_avg << " seconds" << endl;
	      cout << " stdev_for_iter   = " << stdev << " seconds" << endl;
	      cout << " Total time (compute+comm)  =  " << totaltime << ";  computetime = " << computetime << endl;
	      cout << " Usertime/element/iter  = " << (usrtime/dance_steps[k].numelements/dance_steps[k].numiters)*1e6 << " microseconds"<<endl;
	      cout << " Systime/element/iter   = " << (systime/dance_steps[k].numelements/dance_steps[k].numiters)*1e6 << " microseconds"<<endl;
	      cout << " Iters/s            = " << (dance_steps[k].numiters/wallclock) << endl;
	      cout << " ME/s               = " << (dance_steps[k].numiters*dance_steps[k].numelements*dance_steps[k].numprocs)/wallclock/1e6 << endl;
	      cout << " MFLOP/s            = " << (dance_steps[k].numiters*dance_steps[k].numelements*dance_steps[k].numprocs*dance_steps[k].flopsperelement)/wallclock/1e6 << endl;
	      cout << "Done."<<endl;
	    }

	}

    } // for k : dance steps loop

  Time endtime;
  double totalexectime=(double)endtime - (double)starttime;
    cout << " Total execution time is : " << totalexectime << endl;

  pvm_lvgroup("barrier");
  pvm_exit();
  return 0;
  
}
