#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include "pvm3.h"
#include "pvmspawn.h"


int spawnprocs(char *name, int spawntype, int num, int tid[])
{
  int i;
  char where[MAXNPROC][MAXHOSTLEN];
  int rc;
  int numspawned=0;
  int hostavailable;
  int info;
  char *stupid[1];


  switch (parsehostfile(name,num+1,where)) {
  case HFPARSED:
    for (i=1;i<num+1;i++) {
      if (pvm_mstat(where[i])) {
#ifdef DEBUG
	fprintf(stderr, "%s is not in virtual machine...adding it\n",where[i]);
#endif
	stupid[0]=(char *)where[i];
	if ((rc=pvm_addhosts(stupid,1,&info))<1) {
	  fprintf(stderr,"%s could not be added (rc=%d) - process %d will be mapped "
		  " by PVM.\n",where[i],rc,i);
	  hostavailable=0;
	} else {
	  hostavailable=1;
	}
      } else {
	hostavailable=1;
      }
	  
      if (hostavailable) {
#ifdef DEBUG
	fprintf(stderr, "spawning process %d on %s\n",i,where[i]);
#endif
	
	if ((rc=pvm_spawn(name,
			  (char **)NULL,
			  spawntype+PvmTaskHost,
			  where[i],
			  1,
			  &(tid[i])))==1) {
	  numspawned++;
	} 
      } else {
#ifdef DEBUG
	fprintf(stderr,"spawning process %d on unknown machine\n",i);
#endif
	if ((rc=pvm_spawn(name,
			  (char **)NULL,
			  spawntype,
			  "",
			  1,
			  &(tid[i])))==1) {
	  numspawned++;
	} 
      }
    }
    return numspawned;
    break;
  case HFNOHF:
    return (num>0)? pvm_spawn(name,               /* Program */ 
			      (char **)NULL,      /* Arguments */
			      spawntype,          /* Class of where */
			      "",                 /* Where */
			      num,                /* Number of Processes */
			      &(tid[1]))   : 0 ;  /* Returned Thread IDS */ 
    break;
    
  case HFBAD:
  default:
    return HFBAD;
  }
}



void uppercase(char *t)
{
  for (;*t!=0;t++)
    *t=toupper(*t);
}

int parsehostfile(char *execname, int numprocs, char hn[][MAXHOSTLEN])
{
  int i;
  int numdone;
  FILE *inp;
  char buf[BUFLEN];
  char host[MAXHOSTLEN];
  int first;
  int last;
  int lineno;


  strcpy(buf,execname);
  strcat(buf,".hf");

  if ((inp=fopen(buf, "r"))==NULL)
    return HFNOHF;


  for (i=0;i<numprocs;i++)
    hn[i][0]=0;

  
  numdone=0;
  lineno=0;
  while (!feof(inp)) {
    fgets(buf,BUFLEN,inp);
    ++lineno;
#ifdef DEBUG
    fprintf(stderr, "Parsing buffer *%s*\n",buf);
#endif
    if (*buf=='#' || *buf==0 || isspace(*buf))
      continue;

    first=last=0;

    sscanf(buf,"%d-%d %s", &first, &last, host);
/*    uppercase(host); */
    if (last!=0) {
      last=(last>=numprocs) ? numprocs : last;
      if (first<numprocs) { 
	for (i=first;i<=last;i++) {
	  if (hn[i][0]!=0) {
	    fprintf(stderr,"Error in hostfile at line %d: process %d has"
		    " multiple hosts\n",lineno,i);
	    fclose(inp);
	    return HFBAD;
	  } else {
	    strcpy(hn[i],host);
	  }
	}
	continue;
      }
    }
    
    sscanf(buf,"%d %s",&first,host);
/*    uppercase(host); */
    if (first<numprocs) {
      if (hn[first][0]!=0) {
	fprintf(stderr,"Error in hostfile at line %d: process %d has" 
		" multiple hosts\n",lineno,i);
	fclose(inp);
	return HFBAD;
      } else { 
	strcpy(hn[first],host);
      }
    }
  }

  fclose(inp);

  first=0;
  for (i=0;i<numprocs;i++) {
    if (hn[i][0]==0) {
      fprintf(stderr,"Error in hostfile - no host specified for process %d\n",
	      i);
      first=1;
    }
  }

  if (first)
    return HFBAD;

  gethostname(host,MAXHOSTLEN);

  if (strcmp(host,hn[0])) {
    fprintf(stderr,"Warning: hostfile specifies process 0 should be on\n");
    fprintf(stderr,"         %s, but it was started on %s\n", hn[0], host);
  }

  return HFPARSED;


}
    
    
    
  
